# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
from math import pi
from pyomo.common.collections import ComponentMap
from pyomo.contrib.fbbt.interval import (
    BoolFlag,
    eq,
    ineq,
    ranged,
    if_,
    add,
    acos,
    asin,
    atan,
    cos,
    div,
    exp,
    interval_abs,
    log,
    log10,
    mul,
    power,
    sin,
    sub,
    tan,
)
from pyomo.core.base.expression import Expression
from pyomo.core.expr.numeric_expr import (
    NumericExpression,
    NegationExpression,
    ProductExpression,
    DivisionExpression,
    PowExpression,
    AbsExpression,
    UnaryFunctionExpression,
    MonomialTermExpression,
    LinearExpression,
    SumExpression,
    ExternalFunctionExpression,
    Expr_ifExpression,
)
from pyomo.core.expr.logical_expr import BooleanExpression
from pyomo.core.expr.relational_expr import (
    EqualityExpression,
    InequalityExpression,
    RangedExpression,
)
from pyomo.core.expr.numvalue import native_numeric_types, native_types, value
from pyomo.core.expr.visitor import StreamBasedExpressionVisitor
from pyomo.repn.util import BeforeChildDispatcher, ExitNodeDispatcher

inf = float('inf')
logger = logging.getLogger(__name__)


class ExpressionBoundsBeforeChildDispatcher(BeforeChildDispatcher):
    __slots__ = ()

    def __init__(self):
        self[ExternalFunctionExpression] = self._before_external_function

    @staticmethod
    def _before_external_function(visitor, child):
        # [ESJ 10/6/23]: If external functions ever implement callbacks to help with
        # this then this should use them
        return False, (-inf, inf)

    @staticmethod
    def _before_native_numeric(visitor, child):
        return False, (child, child)

    @staticmethod
    def _before_native_logical(visitor, child):
        return False, (BoolFlag(child), BoolFlag(child))

    @staticmethod
    def _before_var(visitor, child):
        leaf_bounds = visitor.leaf_bounds
        if child in leaf_bounds:
            pass
        elif child.is_fixed() and visitor.use_fixed_var_values_as_bounds:
            val = child.value
            try:
                ans = visitor._before_child_handlers[val.__class__](visitor, val)
            except ValueError:
                raise ValueError(
                    "Var '%s' is fixed to None. This value cannot be used to "
                    "calculate bounds." % child.name
                ) from None
            leaf_bounds[child] = ans[1]
            return ans
        else:
            lb = child.lb
            ub = child.ub
            if lb is None:
                lb = -inf
            if ub is None:
                ub = inf
            leaf_bounds[child] = (lb, ub)
        return False, leaf_bounds[child]

    @staticmethod
    def _before_named_expression(visitor, child):
        leaf_bounds = visitor.leaf_bounds
        if child in leaf_bounds:
            return False, leaf_bounds[child]
        else:
            return True, None

    @staticmethod
    def _before_param(visitor, child):
        val = child.value
        return visitor._before_child_handlers[val.__class__](visitor, val)

    @staticmethod
    def _before_string(visitor, child):
        raise ValueError(
            f"{child!r} ({type(child).__name__}) is not a valid numeric type. "
            f"Cannot compute bounds on expression."
        )

    @staticmethod
    def _before_invalid(visitor, child):
        raise ValueError(
            f"{child!r} ({type(child).__name__}) is not a valid numeric type. "
            f"Cannot compute bounds on expression."
        )

    @staticmethod
    def _before_complex(visitor, child):
        raise ValueError(
            f"Cannot compute bounds on expressions containing "
            f"complex numbers. Encountered when processing {child}"
        )

    @staticmethod
    def _before_npv(visitor, child):
        val = value(child)
        return visitor._before_child_handlers[val.__class__](visitor, val)


def _handle_ProductExpression(visitor, node, arg1, arg2):
    if arg1 is arg2:
        return power(*arg1, 2, 2, feasibility_tol=visitor.feasibility_tol)
    return mul(*arg1, *arg2)


def _handle_SumExpression(visitor, node, *args):
    bnds = (0, 0)
    for arg in args:
        bnds = add(*bnds, *arg)
    return bnds


def _handle_DivisionExpression(visitor, node, arg1, arg2):
    return div(*arg1, *arg2, feasibility_tol=visitor.feasibility_tol)


def _handle_PowExpression(visitor, node, arg1, arg2):
    return power(*arg1, *arg2, feasibility_tol=visitor.feasibility_tol)


def _handle_NegationExpression(visitor, node, arg):
    return sub(0, 0, *arg)


def _handle_exp(visitor, node, arg):
    return exp(*arg)


def _handle_log(visitor, node, arg):
    return log(*arg)


def _handle_log10(visitor, node, arg):
    return log10(*arg)


def _handle_sin(visitor, node, arg):
    return sin(*arg)


def _handle_cos(visitor, node, arg):
    return cos(*arg)


def _handle_tan(visitor, node, arg):
    return tan(*arg)


def _handle_asin(visitor, node, arg):
    return asin(*arg, -pi / 2, pi / 2, visitor.feasibility_tol)


def _handle_acos(visitor, node, arg):
    return acos(*arg, 0, pi, visitor.feasibility_tol)


def _handle_atan(visitor, node, arg):
    return atan(*arg, -pi / 2, pi / 2)


def _handle_sqrt(visitor, node, arg):
    return power(*arg, 0.5, 0.5, feasibility_tol=visitor.feasibility_tol)


def _handle_AbsExpression(visitor, node, arg):
    return interval_abs(*arg)


def _handle_UnaryFunctionExpression(visitor, node, arg):
    return _unary_function_dispatcher[node.getname()](visitor, node, arg)


def _handle_named_expression(visitor, node, arg):
    visitor.leaf_bounds[node] = arg
    return arg


def _handle_unknowable_bounds(visitor, node, arg):
    return -inf, inf


def _handle_equality(visitor, node, arg1, arg2):
    return eq(*arg1, *arg2, feasibility_tol=visitor.feasibility_tol)


def _handle_inequality(visitor, node, arg1, arg2):
    return ineq(*arg1, *arg2, feasibility_tol=visitor.feasibility_tol)


def _handle_ranged(visitor, node, arg1, arg2, arg3):
    return ranged(*arg1, *arg2, *arg3, feasibility_tol=visitor.feasibility_tol)


def _handle_expr_if(visitor, node, arg1, arg2, arg3):
    return if_(*arg1, *arg2, *arg3)


_unary_function_dispatcher = {
    'exp': _handle_exp,
    'log': _handle_log,
    'log10': _handle_log10,
    'sin': _handle_sin,
    'cos': _handle_cos,
    'tan': _handle_tan,
    'asin': _handle_asin,
    'acos': _handle_acos,
    'atan': _handle_atan,
    'sqrt': _handle_sqrt,
}


class ExpressionBoundsExitNodeDispatcher(ExitNodeDispatcher):
    def unexpected_expression_type(self, visitor, node, *args):
        if isinstance(node, NumericExpression):
            ans = -inf, inf
        elif isinstance(node, BooleanExpression):
            ans = BoolFlag(False), BoolFlag(True)
        else:
            super().unexpected_expression_type(visitor, node, *args)
        logger.warning(
            f"Unexpected expression node type '{type(node).__name__}' "
            f"found while walking expression tree; returning {ans} "
            "for the expression bounds."
        )
        return ans


class ExpressionBoundsVisitor(StreamBasedExpressionVisitor):
    """
    Walker to calculate bounds on an expression, from leaf to root, with
    caching of terminal node bounds (Vars and Expressions)

    NOTE: If anything changes on the model (e.g., Var bounds, fixing, mutable
    Param values, etc), then you need to either create a new instance of this
    walker, or clear self.leaf_bounds!

    Parameters
    ----------
    leaf_bounds: ComponentMap in which to cache bounds at leaves of the expression
        tree
    feasibility_tol: float, feasibility tolerance for interval arithmetic
        calculations
    use_fixed_var_values_as_bounds: bool, whether or not to use the values of
        fixed Vars as the upper and lower bounds for those Vars or to instead
        ignore fixed status and use the bounds. Set to 'True' if you do not
        anticipate the fixed status of Variables to change for the duration that
        the computed bounds should be valid.
    """

    _before_child_handlers = ExpressionBoundsBeforeChildDispatcher()
    _operator_dispatcher = ExpressionBoundsExitNodeDispatcher(
        {
            ProductExpression: _handle_ProductExpression,
            DivisionExpression: _handle_DivisionExpression,
            PowExpression: _handle_PowExpression,
            AbsExpression: _handle_AbsExpression,
            SumExpression: _handle_SumExpression,
            MonomialTermExpression: _handle_ProductExpression,
            NegationExpression: _handle_NegationExpression,
            UnaryFunctionExpression: _handle_UnaryFunctionExpression,
            LinearExpression: _handle_SumExpression,
            Expression: _handle_named_expression,
            ExternalFunctionExpression: _handle_unknowable_bounds,
            EqualityExpression: _handle_equality,
            InequalityExpression: _handle_inequality,
            RangedExpression: _handle_ranged,
            Expr_ifExpression: _handle_expr_if,
        }
    )

    def __init__(
        self,
        leaf_bounds=None,
        feasibility_tol=1e-8,
        use_fixed_var_values_as_bounds=False,
    ):
        super().__init__()
        self.leaf_bounds = leaf_bounds if leaf_bounds is not None else ComponentMap()
        self.feasibility_tol = feasibility_tol
        self.use_fixed_var_values_as_bounds = use_fixed_var_values_as_bounds

    def initializeWalker(self, expr):
        walk, result = self.beforeChild(None, expr, 0)
        if not walk:
            return False, result
        return True, expr

    def beforeChild(self, node, child, child_idx):
        return self._before_child_handlers[child.__class__](self, child)

    def exitNode(self, node, data):
        return self._operator_dispatcher[node.__class__](self, node, *data)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import sys
from weakref import ref as weakref_ref

from pyomo.common.deprecation import deprecated, RenamedClass
from pyomo.common.formatting import tabular_writer
from pyomo.common.log import is_debug_set
from pyomo.common.modeling import NOTSET
from pyomo.common.numeric_types import value
from pyomo.common.timing import ConstructionTimer
from pyomo.core.expr.numvalue import NumericValue
from pyomo.core.base.component import ComponentData, ModelComponentFactory
from pyomo.core.base.global_set import UnindexedComponent_index
from pyomo.core.base.indexed_component import IndexedComponent
from pyomo.core.base.misc import apply_indexed_rule

logger = logging.getLogger('pyomo.core')


class ConnectorData(ComponentData, NumericValue):
    """Holds the actual connector information"""

    __slots__ = ('vars', 'aggregators')

    def __init__(self, component=None):
        """Constructor"""
        #
        # These lines represent in-lining of the
        # following constructors:
        #   - ComponentData
        #   - NumericValue
        self._component = weakref_ref(component) if (component is not None) else None
        self._index = NOTSET

        self.vars = {}
        self.aggregators = {}

    def set_value(self, value):
        msg = "Cannot specify the value of a connector '%s'"
        raise ValueError(msg % self.name)

    def is_fixed(self):
        """Return True if all vars/expressions in the Connector are fixed"""
        return all(v.is_fixed() for v in self._iter_vars())

    def is_constant(self):
        """Return False

        Because the expression generation logic will attempt to evaluate
        constant subexpressions, a Connector can never be constant.
        """
        return False

    def is_potentially_variable(self):
        """Return True as connectors may (should!) contain variables"""
        return True

    def polynomial_degree(self):
        ans = 0
        for v in self._iter_vars():
            tmp = v.polynomial_degree()
            if tmp is None:
                return None
            ans = max(ans, tmp)
        return ans

    def is_binary(self):
        return len(self) and all(v.is_binary() for v in self._iter_vars())

    def is_integer(self):
        return len(self) and all(v.is_integer() for v in self._iter_vars())

    def is_continuous(self):
        return len(self) and all(v.is_continuous() for v in self._iter_vars())

    def add(self, var, name=None, aggregate=None):
        if name is None:
            name = var.local_name
        if name in self.vars:
            raise ValueError(
                "Cannot insert duplicate variable name "
                "'%s' into Connector '%s'" % (name, self.name)
            )
        self.vars[name] = var
        if aggregate is not None:
            self.aggregators[name] = aggregate

    def _iter_vars(self):
        for var in self.vars.values():
            if not hasattr(var, 'is_indexed') or not var.is_indexed():
                yield var
            else:
                for v in var.values():
                    yield v


class _ConnectorData(metaclass=RenamedClass):
    __renamed__new_class__ = ConnectorData
    __renamed__version__ = '6.7.2'


@ModelComponentFactory.register(
    "A bundle of variables that can be manipulated together."
)
@deprecated(
    "Use of pyomo.connectors is deprecated. "
    "Its functionality has been replaced by pyomo.network.",
    version='5.6.9',
)
class Connector(IndexedComponent):
    """A collection of variables, which may be defined over a index

    The idea behind a Connector is to create a bundle of variables that
    can be manipulated as a single variable within constraints.  While
    Connectors inherit from variable (mostly so that the expression
    infrastructure can manipulate them), they are not actual variables
    that are exposed to the solver.  Instead, a preprocessor
    (ConnectorExpander) will look for expressions that involve
    connectors and replace the single constraint with a list of
    constraints that involve the original variables contained within the
    Connector.

    Parameters
    ----------
    name : str
        The name of this connector

    index
        The index set that defines the distinct connectors.  By default,
        this is None, indicating that there is a single connector.

    """

    def __new__(cls, *args, **kwds):
        if cls != Connector:
            return super(Connector, cls).__new__(cls)
        if args == ():
            return ScalarConnector.__new__(ScalarConnector)
        else:
            return IndexedConnector.__new__(IndexedConnector)

    # TODO: default keyword is  not used?  Need to talk to Bill ...?
    def __init__(self, *args, **kwd):
        kwd.setdefault('ctype', Connector)
        self._rule = kwd.pop('rule', None)
        self._initialize = kwd.pop('initialize', {})
        self._implicit = kwd.pop('implicit', {})
        self._extends = kwd.pop('extends', None)
        IndexedComponent.__init__(self, *args, **kwd)
        self._conval = {}

    #
    # This method must be defined on subclasses of
    # IndexedComponent
    #
    def _getitem_when_not_present(self, idx):
        _conval = self._data[idx] = ConnectorData(component=self)
        return _conval

    def construct(self, data=None):
        if is_debug_set(logger):  # pragma:nocover
            logger.debug(
                "Constructing Connector, name=%s, from data=%s" % (self.name, data)
            )
        if self._constructed:
            return
        timer = ConstructionTimer(self)
        self._constructed = True
        #
        # Construct ConnectorData objects for all index values
        #
        if self.is_indexed():
            self._initialize_members(self._index_set)
        else:
            self._data[None] = self
            self._initialize_members([None])
        timer.report()

    def _initialize_members(self, initSet):
        for idx in initSet:
            tmp = self[idx]
            for key in self._implicit:
                tmp.add(None, key)
            if self._extends:
                for key, val in self._extends.vars.items():
                    tmp.add(val, key)
            for key, val in self._initialize.items():
                tmp.add(val, key)
            if self._rule:
                items = apply_indexed_rule(self, self._rule, self._parent(), idx)
                for key, val in items.items():
                    tmp.add(val, key)

    def _pprint(self):
        """Print component information."""

        def _line_generator(k, v):
            for _k, _v in sorted(v.vars.items()):
                if _v is None:
                    _len = '-'
                elif _k in v.aggregators:
                    _len = '*'
                elif hasattr(_v, '__len__'):
                    _len = len(_v)
                else:
                    _len = 1
                yield _k, _len, str(_v)

        return (
            [
                ("Size", len(self)),
                ("Index", self._index_set if self.is_indexed() else None),
            ],
            self.items,
            ("Name", "Size", "Variable"),
            _line_generator,
        )

    def display(self, prefix="", ostream=None):
        """
        Print component state information

        This duplicates logic in Component.pprint()
        """
        if not self.active:
            return
        if ostream is None:
            ostream = sys.stdout
        tab = "    "
        ostream.write(prefix + self.local_name + " : ")
        ostream.write("Size=" + str(len(self)))

        ostream.write("\n")

        def _line_generator(k, v):
            for _k, _v in sorted(v.vars.items()):
                if _v is None:
                    _val = '-'
                elif not hasattr(_v, 'is_indexed') or not _v.is_indexed():
                    _val = str(value(_v))
                else:
                    _val = "{%s}" % (
                        ', '.join(
                            '%r: %r' % (x, value(_v[x])) for x in sorted(_v._data)
                        ),
                    )
                yield _k, _val

        tabular_writer(
            ostream,
            prefix + tab,
            ((k, v) for k, v in self._data.items()),
            ("Name", "Value"),
            _line_generator,
        )


class ScalarConnector(Connector, ConnectorData):
    def __init__(self, *args, **kwd):
        ConnectorData.__init__(self, component=self)
        Connector.__init__(self, *args, **kwd)
        self._index = UnindexedComponent_index


class SimpleConnector(metaclass=RenamedClass):
    __renamed__new_class__ = ScalarConnector
    __renamed__version__ = '6.0'


class IndexedConnector(Connector):
    """An array of connectors"""

    pass



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import sys
import time
from pyomo.common.download import FileDownloader, DownloadFactory
from pyomo.scripting.pyomo_parser import add_subparser

MAX_RETRIES = 3
RETRY_SLEEP_SECONDS = 5


class GroupDownloader:
    def __init__(self):
        self.downloader = FileDownloader()

    def create_parser(self, parser):
        return self.downloader.create_parser(parser)

    def call(self, args, unparsed):
        logger = logging.getLogger('pyomo.common')
        original_level = logger.level
        logger.setLevel(logging.INFO)
        try:
            return self._call_impl(args, unparsed, logger)
        finally:
            logger.setLevel(original_level)

    def _call_impl(self, args, unparsed, logger):
        results = []
        result_fmt = "[%s]  %s"
        returncode = 0

        self.downloader.cacert = args.cacert
        self.downloader.insecure = args.insecure

        logger.info(
            "As of February 9, 2023, AMPL GSL can no longer be downloaded "
            "through download-extensions. Visit https://portal.ampl.com/ "
            "to download the AMPL GSL binaries."
        )

        for target in DownloadFactory:
            attempt = 0
            result = "FAIL"

            while attempt < MAX_RETRIES:
                try:
                    ext = DownloadFactory(target, downloader=self.downloader)

                    if hasattr(ext, "skip") and ext.skip():
                        result = "SKIP"
                    elif hasattr(ext, "__call__"):
                        ext()
                        result = " OK "
                    else:
                        result = " OK "

                    break

                except SystemExit:
                    _info = sys.exc_info()
                    _cls = (
                        f"{_info[0].__name__ if _info[0] is not None else 'NoneType'}: "
                    )
                    logger.error(_cls + str(_info[1]))
                    if attempt + 1 == MAX_RETRIES:
                        returncode |= 2
                except Exception:
                    _info = sys.exc_info()
                    _cls = (
                        f"{_info[0].__name__ if _info[0] is not None else 'NoneType'}: "
                    )
                    logger.error(_cls + str(_info[1]))
                    if attempt + 1 == MAX_RETRIES:
                        returncode |= 1
                finally:
                    if result != " OK ":
                        attempt += 1
                        if attempt < MAX_RETRIES:
                            logger.info(
                                f"Retrying download of '{target}' "
                                f"(attempt {attempt + 1} of {MAX_RETRIES})"
                            )
                            time.sleep(RETRY_SLEEP_SECONDS)

            results.append(result_fmt % (result, target))

        logger.info("Finished downloading Pyomo extensions.")
        logger.info(
            "The following extensions were downloaded:\n    " + "\n    ".join(results)
        )
        return returncode


#
# Add a subparser for the download-extensions command
#
_group_downloader = GroupDownloader()
_parser = _group_downloader.create_parser(
    add_subparser(
        'download-extensions',
        func=_group_downloader.call,
        help='Download compiled extension modules',
        add_help=False,
        description='This downloads all registered (compiled) extension modules',
    )
)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import collections.abc
import pickle
from io import StringIO

import pyomo.common.unittest as unittest
import pyomo.kernel as pmo
from pyomo.common.log import LoggingIntercept
from pyomo.common.tee import capture_output
from pyomo.core.kernel.base import ICategorizedObject, ICategorizedObjectContainer
from pyomo.core.kernel.homogeneous_container import IHomogeneousContainer
from pyomo.core.kernel.dict_container import DictContainer
from pyomo.core.kernel.block import block, block_dict

#
# There are no fully implemented test suites in this
# file. These classes are meant to be used to test
# full implementations of DictContainer containers.
# To test a DictContainer implementation, just import
# this class and use it as a subclass in another test suite.
#

# Note: we need to test with a pickle protocol
#       that is high enough to support __slots__
#       and weakref (bas
_pickle_test_protocol = pickle.HIGHEST_PROTOCOL


class _bad_ctype:
    ctype = "_this_is_definitely_not_the_ctype_being_tested"


class _TestDictContainerBase:
    # set by derived class
    _container_type = None
    _ctype_factory = None

    def test_overwrite_warning(self):
        c = self._container_type()
        out = StringIO()
        with LoggingIntercept(out, 'pyomo.core'):
            c[0] = self._ctype_factory()
            c[0] = c[0]
        assert out.getvalue() == ""
        with LoggingIntercept(out, 'pyomo.core'):
            c[0] = self._ctype_factory()
        assert out.getvalue() == (
            "Implicitly replacing the entry [0] "
            "(type=%s) with a new object (type=%s). "
            "This is usually indicative of a modeling "
            "error. To avoid this warning, delete the "
            "original object from the container before "
            "assigning a new object.\n"
            % (
                self._ctype_factory().__class__.__name__,
                self._ctype_factory().__class__.__name__,
            )
        )

    def test_pprint(self):
        import pyomo.kernel

        # Not really testing what the output is, just that
        # an error does not occur. The pprint functionality
        # is still in the early stages.
        with capture_output() as OUT:
            cdict = self._container_type({None: self._ctype_factory()})
            pyomo.kernel.pprint(cdict)
            b = block()
            b.cdict = cdict
            pyomo.kernel.pprint(cdict)
            pyomo.kernel.pprint(b)
            m = block()
            m.b = b
            pyomo.kernel.pprint(cdict)
            pyomo.kernel.pprint(b)
            pyomo.kernel.pprint(m)

    def test_ctype(self):
        c = self._container_type()
        ctype = self._ctype_factory().ctype
        self.assertIs(c.ctype, ctype)
        self.assertIs(type(c)._ctype, ctype)
        self.assertIs(self._container_type._ctype, ctype)

    def test_init1(self):
        cdict = self._container_type()

    def test_init2(self):
        index = ['a', 1, None, (1,), (1, 2)]
        self._container_type((i, self._ctype_factory()) for i in index)
        self._container_type(((i, self._ctype_factory()) for i in index))
        with self.assertRaises(TypeError):
            self._container_type(*tuple((i, self._ctype_factory()) for i in index))
        c = self._container_type(a=self._ctype_factory(), b=self._ctype_factory())
        self.assertEqual(len(c), 2)
        self.assertTrue('a' in c)
        self.assertTrue('b' in c)

    def test_ordered_init(self):
        cdict = self._container_type()
        cdict[None] = self._ctype_factory()
        cdict[1] = self._ctype_factory()
        cdict['a'] = self._ctype_factory()
        del cdict[None]
        cdict[2] = self._ctype_factory()
        cdict[3] = self._ctype_factory()
        del cdict[3]
        cdict[None] = self._ctype_factory()
        cdict[-1] = self._ctype_factory()
        cdict['bc'] = self._ctype_factory()
        cdict[3] = self._ctype_factory()
        self.assertEqual(list(cdict.keys()), [1, 'a', 2, None, -1, 'bc', 3])

    def test_type(self):
        cdict = self._container_type()
        self.assertTrue(isinstance(cdict, ICategorizedObject))
        self.assertTrue(isinstance(cdict, ICategorizedObjectContainer))
        self.assertTrue(isinstance(cdict, IHomogeneousContainer))
        self.assertTrue(isinstance(cdict, DictContainer))
        self.assertTrue(isinstance(cdict, collections.abc.Mapping))
        self.assertTrue(isinstance(cdict, collections.abc.MutableMapping))
        self.assertTrue(issubclass(type(cdict), collections.abc.Mapping))
        self.assertTrue(issubclass(type(cdict), collections.abc.MutableMapping))

    def test_len1(self):
        cdict = self._container_type()
        self.assertEqual(len(cdict), 0)

    def test_len2(self):
        index = ['a', 1, None, (1,), (1, 2)]
        cdict = self._container_type((i, self._ctype_factory()) for i in index)
        self.assertEqual(len(cdict), len(index))

    def test_setitem(self):
        cdict = self._container_type()
        index = ['a', 1, None, (1,), (1, 2)]
        for i in index:
            self.assertTrue(i not in cdict)
        for cnt, i in enumerate(index, 1):
            cdict[i] = self._ctype_factory()
            self.assertEqual(len(cdict), cnt)
            self.assertTrue(i in cdict)

    # The immediately following this one was originally written when
    # implicit assignment for update was supported. This should be
    # examined more carefully before supporting it.
    # For now just test that implicit assignment raises an exception
    def test_wrong_type_init(self):
        index = ['a', 1, None, (1,), (1, 2)]
        with self.assertRaises(TypeError):
            c = self._container_type((i, _bad_ctype()) for i in index)

    def test_wrong_type_update(self):
        index = ['a', 1, None, (1,), (1, 2)]
        c = self._container_type()
        with self.assertRaises(TypeError):
            c.update((i, _bad_ctype()) for i in index)

    def test_wrong_type_setitem(self):
        c = self._container_type()
        with self.assertRaises(TypeError):
            c[1] = _bad_ctype()
        c[1] = self._ctype_factory()
        with self.assertRaises(TypeError):
            c[1] = _bad_ctype()

    def test_has_parent_init(self):
        c = self._container_type()
        c[1] = self._ctype_factory()
        with self.assertRaises(ValueError):
            d = self._container_type(c)
        with self.assertRaises(ValueError):
            d = self._container_type(dict(c))

    def test_has_parent_update(self):
        c = self._container_type()
        c[1] = self._ctype_factory()
        c.update(c)
        self.assertEqual(len(c), 1)
        c.update(dict(c))
        self.assertEqual(len(c), 1)
        d = self._container_type()
        with self.assertRaises(ValueError):
            d.update(c)
        with self.assertRaises(ValueError):
            d.update(dict(c))

    def test_has_parent_setitem(self):
        c = self._container_type()
        c[1] = self._ctype_factory()
        c[1] = c[1]
        with self.assertRaises(ValueError):
            c[2] = c[1]
        d = self._container_type()
        with self.assertRaises(ValueError):
            d[None] = c[1]

    # make sure an existing Data object IS replaced
    # by a call to setitem and not simply updated.
    def test_setitem_exists_overwrite(self):
        index = ['a', 1, None, (1,), (1, 2)]
        c = self._container_type((i, self._ctype_factory()) for i in index)
        self.assertEqual(len(c), len(index))
        for i in index:
            self.assertTrue(i in c)
            cdata = c[i]
            c[i] = self._ctype_factory()
            self.assertEqual(len(c), len(index))
            self.assertTrue(i in c)
            self.assertNotEqual(id(cdata), id(c[i]))
            self.assertEqual(cdata.parent, None)

    def test_delitem(self):
        index = ['a', 1, None, (1,), (1, 2)]
        c = self._container_type((i, self._ctype_factory()) for i in index)
        self.assertEqual(len(c), len(index))
        for cnt, i in enumerate(index, 1):
            self.assertTrue(i in c)
            cdata = c[i]
            self.assertEqual(id(cdata.parent), id(c))
            del c[i]
            self.assertEqual(len(c), len(index) - cnt)
            self.assertTrue(i not in c)
            self.assertEqual(cdata.parent, None)

    def test_iter(self):
        index = ['a', 1, None, (1,), (1, 2)]
        c = self._container_type((i, self._ctype_factory()) for i in index)
        self.assertEqual(len(c), len(index))
        comp_index = [i for i in c]
        self.assertEqual(len(comp_index), len(index))
        for idx in comp_index:
            self.assertTrue(idx in index)

    def test_pickle(self):
        index = ['a', 1, None, (1,), (1, 2)]
        cdict = self._container_type((i, self._ctype_factory()) for i in index)
        cdict[0] = self._container_type()
        index.append(0)
        for i in index:
            self.assertTrue(cdict[i].parent is cdict)
        pickled_cdict = pickle.loads(pickle.dumps(cdict))
        self.assertTrue(isinstance(pickled_cdict, self._container_type))
        self.assertTrue(pickled_cdict.parent is None)
        self.assertEqual(len(pickled_cdict), len(index))
        self.assertNotEqual(id(pickled_cdict), id(cdict))
        for i in index:
            self.assertNotEqual(id(pickled_cdict[i]), id(cdict[i]))
            self.assertTrue(pickled_cdict[i].parent is pickled_cdict)
            self.assertTrue(cdict[i].parent is cdict)

    def test_keys(self):
        index = ['a', 1, None, (1,), (1, 2)]
        raw_constraint_dict = {i: self._ctype_factory() for i in index}
        c = self._container_type(raw_constraint_dict)
        self.assertEqual(
            sorted(list(raw_constraint_dict.keys()), key=str),
            sorted(list(c.keys()), key=str),
        )

    def test_values(self):
        index = ['a', 1, None, (1,), (1, 2)]
        raw_constraint_dict = {i: self._ctype_factory() for i in index}
        c = self._container_type(raw_constraint_dict)
        self.assertEqual(
            sorted(list(id(_v) for _v in raw_constraint_dict.values()), key=str),
            sorted(list(id(_v) for _v in c.values()), key=str),
        )

    def test_items(self):
        index = ['a', 1, None, (1,), (1, 2)]
        raw_constraint_dict = {i: self._ctype_factory() for i in index}
        c = self._container_type(raw_constraint_dict)
        self.assertEqual(
            sorted(
                list((_i, id(_v)) for _i, _v in raw_constraint_dict.items()), key=str
            ),
            sorted(list((_i, id(_v)) for _i, _v in c.items()), key=str),
        )

    def test_update(self):
        index = ['a', 1, None, (1,), (1, 2)]
        raw_constraint_dict = {i: self._ctype_factory() for i in index}
        c = self._container_type()
        c.update(raw_constraint_dict)
        self.assertEqual(
            sorted(list(raw_constraint_dict.keys()), key=str),
            sorted(list(c.keys()), key=str),
        )

    def test_clear(self):
        c = self._container_type()
        c[1] = self._ctype_factory()
        c1 = c[1]
        with self.assertRaises(ValueError):
            c[None] = c1
        d = self._container_type()
        with self.assertRaises(ValueError):
            d[1] = c1
        c.clear()
        d[1] = c1

    def test_eq(self):
        cdict1 = self._container_type()
        cdict1[1] = self._ctype_factory()
        cdict2 = self._container_type()
        cdict2[1] = self._ctype_factory()

        self.assertNotEqual(cdict1, set())
        self.assertFalse(cdict1 == set())
        self.assertNotEqual(cdict1, list())
        self.assertFalse(cdict1 == list())
        self.assertNotEqual(cdict1, tuple())
        self.assertFalse(cdict1 == tuple())
        self.assertNotEqual(cdict1, dict())
        self.assertFalse(cdict1 == dict())

        self.assertTrue(cdict1 == cdict1)
        self.assertEqual(cdict1, cdict1)
        self.assertTrue(cdict1 == dict(cdict1))
        self.assertEqual(cdict1, dict(cdict1))
        self.assertTrue(dict(cdict1) == cdict1)
        self.assertEqual(dict(cdict1), cdict1)

        self.assertFalse(cdict2 == cdict1)
        self.assertTrue(cdict2 != cdict1)
        self.assertNotEqual(cdict2, cdict1)

        self.assertFalse(cdict1 == cdict2)
        self.assertTrue(cdict1 != cdict2)
        self.assertNotEqual(cdict1, cdict2)

        cdict1.clear()

        self.assertNotEqual(cdict1, set())
        self.assertFalse(cdict1 == set())
        self.assertNotEqual(cdict1, list())
        self.assertFalse(cdict1 == list())
        self.assertNotEqual(cdict1, tuple())
        self.assertFalse(cdict1 == tuple())
        self.assertEqual(cdict1, dict())
        self.assertTrue(cdict1 == dict())

    def test_child(self):
        cdict = self._container_type()
        c = self._ctype_factory()
        with self.assertRaises(KeyError):
            cdict.child(1)
        cdict[1] = c
        self.assertIs(cdict.child(1), c)

    def test_name(self):
        children = {}
        children['a'] = self._ctype_factory()
        children[1] = self._ctype_factory()
        children[None] = self._ctype_factory()
        children[(1,)] = self._ctype_factory()
        children[(1, 2)] = self._ctype_factory()
        children['(1,2)'] = self._ctype_factory()
        children['x'] = self._container_type()
        children['x']['y'] = self._ctype_factory()

        for key, c in children.items():
            self.assertTrue(c.parent is None)
            self.assertEqual(c.local_name, None)
            self.assertEqual(c.name, None)

        cdict = self._container_type()
        self.assertTrue(cdict.parent is None)
        self.assertEqual(cdict.local_name, None)
        self.assertEqual(cdict.name, None)
        cdict.update(children)
        names = pmo.generate_names(cdict)
        for key, c in children.items():
            self.assertTrue(c.parent is cdict)
            self.assertEqual(
                c.getname(fully_qualified=False, convert=str), "[%s]" % (str(key))
            )
            self.assertEqual(
                c.getname(fully_qualified=False, convert=repr), "[%s]" % (repr(key))
            )
            self.assertEqual(
                c.getname(fully_qualified=True, convert=str), "[%s]" % (str(key))
            )
            self.assertEqual(
                c.getname(fully_qualified=True, convert=repr), "[%s]" % (repr(key))
            )
            self.assertEqual(c.name, names[c])
        for c in cdict.components():
            self.assertNotEqual(c.name, None)
            self.assertEqual(c.name, names[c])

        model = block()
        model.cdict = cdict
        self.assertTrue(model.parent is None)
        self.assertTrue(cdict.parent is model)
        self.assertEqual(cdict.local_name, "cdict")
        self.assertEqual(cdict.name, "cdict")
        names = pmo.generate_names(model)
        for key, c in children.items():
            self.assertTrue(c.parent is cdict)
            self.assertEqual(
                c.getname(fully_qualified=False, convert=str), "[%s]" % (str(key))
            )
            self.assertEqual(
                c.getname(fully_qualified=False, convert=repr), "[%s]" % (repr(key))
            )
            self.assertEqual(
                c.getname(fully_qualified=True, convert=str), "cdict[%s]" % (str(key))
            )
            self.assertEqual(
                c.getname(fully_qualified=True, convert=repr), "cdict[%s]" % (repr(key))
            )
            self.assertEqual(c.name, names[c])
        for c in cdict.components():
            self.assertNotEqual(c.name, None)
            self.assertEqual(c.name, names[c])

        b = block()
        b.model = model
        self.assertTrue(b.parent is None)
        self.assertTrue(model.parent is b)
        self.assertTrue(cdict.parent is model)
        self.assertEqual(cdict.local_name, "cdict")
        self.assertEqual(cdict.name, "model.cdict")
        names = pmo.generate_names(b)
        for key, c in children.items():
            self.assertTrue(c.parent is cdict)
            self.assertEqual(
                c.getname(fully_qualified=False, convert=str), "[%s]" % (str(key))
            )
            self.assertEqual(
                c.getname(fully_qualified=False, convert=repr), "[%s]" % (repr(key))
            )
            self.assertEqual(
                c.getname(fully_qualified=True, convert=str),
                "model.cdict[%s]" % (str(key)),
            )
            self.assertEqual(
                c.getname(fully_qualified=True, convert=repr),
                "model.cdict[%s]" % (repr(key)),
            )
            self.assertEqual(c.name, names[c])
        for c in cdict.components():
            self.assertNotEqual(c.name, None)
            self.assertEqual(c.name, names[c])

        bdict = block_dict()
        bdict[0] = b
        self.assertTrue(bdict.parent is None)
        self.assertTrue(b.parent is bdict)
        self.assertTrue(model.parent is b)
        self.assertTrue(cdict.parent is model)
        self.assertEqual(cdict.local_name, "cdict")
        self.assertEqual(cdict.name, "[0].model.cdict")
        for key, c in children.items():
            self.assertTrue(c.parent is cdict)
            self.assertEqual(
                c.getname(fully_qualified=False, convert=str), "[%s]" % (str(key))
            )
            self.assertEqual(
                c.getname(fully_qualified=False, convert=repr), "[%s]" % (repr(key))
            )
            self.assertEqual(
                c.getname(fully_qualified=True, convert=str),
                "[0].model.cdict[%s]" % (str(key)),
            )
            self.assertEqual(
                c.getname(fully_qualified=True, convert=repr),
                "[0].model.cdict[%s]" % (repr(key)),
            )

        m = block()
        m.bdict = bdict
        self.assertTrue(m.parent is None)
        self.assertTrue(bdict.parent is m)
        self.assertTrue(b.parent is bdict)
        self.assertTrue(model.parent is b)
        self.assertTrue(cdict.parent is model)
        self.assertEqual(cdict.local_name, "cdict")
        self.assertEqual(cdict.name, "bdict[0].model.cdict")
        names = pmo.generate_names(m)
        for key, c in children.items():
            self.assertTrue(c.parent is cdict)
            self.assertEqual(
                c.getname(fully_qualified=False, convert=str), "[%s]" % (str(key))
            )
            self.assertEqual(
                c.getname(fully_qualified=False, convert=repr), "[%s]" % (repr(key))
            )
            self.assertEqual(
                c.getname(fully_qualified=True, convert=str),
                "bdict[0].model.cdict[%s]" % (str(key)),
            )
            self.assertEqual(
                c.getname(fully_qualified=True, convert=repr),
                "bdict[0].model.cdict[%s]" % (repr(key)),
            )
            self.assertEqual(c.name, names[c])
        for c in cdict.components():
            self.assertNotEqual(c.name, None)
            self.assertEqual(c.name, names[c])
        names = pmo.generate_names(m)
        for c in m.children():
            self.assertEqual(c.name, names[c])

    def test_preorder_traversal(self):
        traversal = []
        cdict = self._container_type()
        traversal.append(cdict)
        cdict[0] = self._ctype_factory()
        traversal.append(cdict[0])
        cdict[1] = self._container_type()
        traversal.append(cdict[1])
        cdict[1][0] = self._ctype_factory()
        traversal.append(cdict[1][0])
        cdict[2] = self._ctype_factory()
        traversal.append(cdict[2])

        descend = lambda x: not x._is_heterogeneous_container

        self.assertEqual(
            [c.name for c in traversal],
            [c.name for c in pmo.preorder_traversal(cdict, descend=descend)],
        )
        self.assertEqual(
            [id(c) for c in traversal],
            [id(c) for c in pmo.preorder_traversal(cdict, descend=descend)],
        )
        return cdict, traversal

    def test_preorder_traversal_descend_check(self):
        traversal = []
        cdict = self._container_type()
        traversal.append(cdict)
        cdict[0] = self._ctype_factory()
        traversal.append(cdict[0])
        cdict[1] = self._container_type()
        traversal.append(cdict[1])
        cdict[1][0] = self._ctype_factory()
        traversal.append(cdict[1][0])
        cdict[2] = self._ctype_factory()
        traversal.append(cdict[2])

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return False

        descend.seen = []
        order = list(pmo.preorder_traversal(cdict, descend=descend))
        self.assertEqual(len(order), 1)
        self.assertIs(order[0], cdict)
        self.assertEqual(len(descend.seen), 1)
        self.assertIs(descend.seen[0], cdict)

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return not x._is_heterogeneous_container

        descend.seen = []
        order = list(pmo.preorder_traversal(cdict, descend=descend))
        self.assertEqual([c.name for c in traversal], [c.name for c in order])
        self.assertEqual([id(c) for c in traversal], [id(c) for c in order])
        self.assertEqual(
            [c.name for c in traversal if c._is_container],
            [c.name for c in descend.seen],
        )
        self.assertEqual(
            [id(c) for c in traversal if c._is_container], [id(c) for c in descend.seen]
        )

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return not x._is_heterogeneous_container

        descend.seen = []
        order = list(pmo.preorder_traversal(cdict, descend=descend))
        self.assertEqual([c.name for c in traversal], [c.name for c in order])
        self.assertEqual([id(c) for c in traversal], [id(c) for c in order])
        self.assertEqual(
            [c.name for c in traversal if c._is_container],
            [c.name for c in descend.seen],
        )
        self.assertEqual(
            [id(c) for c in traversal if c._is_container], [id(c) for c in descend.seen]
        )
        return cdict, traversal


class _TestActiveDictContainerBase(_TestDictContainerBase):
    def test_active_type(self):
        cdict = self._container_type()
        self.assertTrue(isinstance(cdict, ICategorizedObject))
        self.assertTrue(isinstance(cdict, ICategorizedObjectContainer))
        self.assertTrue(isinstance(cdict, IHomogeneousContainer))
        self.assertTrue(isinstance(cdict, DictContainer))
        self.assertTrue(isinstance(cdict, collections.abc.Mapping))
        self.assertTrue(isinstance(cdict, collections.abc.MutableMapping))

    def test_active(self):
        children = {}
        children['a'] = self._ctype_factory()
        children[1] = self._ctype_factory()
        children[None] = self._ctype_factory()
        children[(1,)] = self._ctype_factory()
        children[(1, 2)] = self._ctype_factory()
        children['(1,2)'] = self._ctype_factory()

        cdict = self._container_type()
        cdict.update(children)
        with self.assertRaises(AttributeError):
            cdict.active = False
        for c in cdict.values():
            with self.assertRaises(AttributeError):
                c.active = False

        model = block()
        model.cdict = cdict
        b = block()
        b.model = model
        bdict = block_dict()
        bdict[0] = b
        bdict[None] = block()
        m = block()
        m.bdict = bdict

        self.assertEqual(m.active, True)
        self.assertEqual(bdict.active, True)
        self.assertEqual(bdict[None].active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(model.active, True)
        self.assertEqual(cdict.active, True)
        for c in cdict.values():
            self.assertEqual(c.active, True)
        for c in cdict.components():
            self.assertEqual(c.active, True)
        for c in cdict.components(active=True):
            self.assertEqual(c.active, True)
        self.assertEqual(len(list(cdict.components())), len(cdict))
        self.assertEqual(
            len(list(cdict.components())), len(list(cdict.components(active=True)))
        )

        m.deactivate(shallow=False)

        self.assertEqual(m.active, False)
        self.assertEqual(bdict.active, False)
        self.assertEqual(bdict[None].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(cdict.active, False)
        for c in cdict.values():
            self.assertEqual(c.active, False)
        self.assertNotEqual(
            len(list(cdict.components())), len(list(cdict.components(active=None)))
        )
        self.assertEqual(len(list(cdict.components(active=True))), 0)

        test_key = list(children.keys())[0]
        del cdict[test_key]
        cdict[test_key] = children[test_key]

        self.assertEqual(m.active, False)
        self.assertEqual(bdict.active, False)
        self.assertEqual(bdict[None].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(cdict.active, False)
        for c in cdict.values():
            self.assertEqual(c.active, False)
        self.assertNotEqual(
            len(list(cdict.components())), len(list(cdict.components(active=None)))
        )
        self.assertEqual(len(list(cdict.components(active=True))), 0)

        del cdict[test_key]
        children[test_key].activate()
        self.assertEqual(children[test_key].active, True)
        cdict[test_key] = children[test_key]

        self.assertEqual(m.active, False)
        self.assertEqual(bdict.active, False)
        self.assertEqual(bdict[None].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(cdict.active, False)
        cdict.activate()
        self.assertEqual(m.active, False)
        self.assertEqual(bdict.active, False)
        self.assertEqual(bdict[None].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(cdict.active, True)
        for key, c in cdict.items():
            if key == test_key:
                self.assertEqual(c.active, True)
            else:
                self.assertEqual(c.active, False)
        for c in cdict.components():
            if c.storage_key == test_key:
                self.assertEqual(c.active, True)
            else:
                self.assertEqual(c.active, False)
        for c in cdict.components(active=True):
            self.assertEqual(c.active, True)
        self.assertNotEqual(
            len(list(cdict.components())), len(list(cdict.components(active=None)))
        )
        self.assertEqual(len(list(cdict.components(active=True))), 1)

        cdict.deactivate()
        m.activate(shallow=False)

        self.assertEqual(m.active, True)
        self.assertEqual(bdict.active, True)
        self.assertEqual(bdict[None].active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(model.active, True)
        self.assertEqual(cdict.active, True)
        for c in cdict.values():
            self.assertEqual(c.active, True)
        for c in cdict.components():
            self.assertEqual(c.active, True)
        for c in cdict.components(active=True):
            self.assertEqual(c.active, True)
        self.assertEqual(len(list(cdict.components())), len(cdict))
        self.assertEqual(
            len(list(cdict.components())), len(list(cdict.components(active=True)))
        )

        cdict.deactivate(shallow=False)

        self.assertEqual(m.active, True)
        self.assertEqual(bdict.active, True)
        self.assertEqual(bdict[None].active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(model.active, True)
        self.assertEqual(cdict.active, False)
        for c in cdict.values():
            self.assertEqual(c.active, False)
        self.assertNotEqual(
            len(list(cdict.components())), len(list(cdict.components(active=None)))
        )
        self.assertEqual(len(list(cdict.components(active=True))), 0)

        cdict.activate(shallow=False)

        self.assertEqual(m.active, True)
        self.assertEqual(bdict.active, True)
        self.assertEqual(bdict[None].active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(model.active, True)
        self.assertEqual(cdict.active, True)
        for c in cdict.values():
            self.assertEqual(c.active, True)
        for c in cdict.components():
            self.assertEqual(c.active, True)
        for c in cdict.components(active=True):
            self.assertEqual(c.active, True)
        self.assertEqual(len(list(cdict.components())), len(cdict))
        self.assertEqual(
            len(list(cdict.components())), len(list(cdict.components(active=True)))
        )

        cdict.deactivate(shallow=False)
        cdict[test_key].activate()

        self.assertEqual(m.active, True)
        self.assertEqual(bdict.active, True)
        self.assertEqual(bdict[None].active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(model.active, True)
        self.assertEqual(cdict.active, False)
        cdict.activate()
        self.assertEqual(m.active, True)
        self.assertEqual(bdict.active, True)
        self.assertEqual(bdict[None].active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(model.active, True)
        self.assertEqual(cdict.active, True)
        for key, c in cdict.items():
            if key == test_key:
                self.assertEqual(c.active, True)
            else:
                self.assertEqual(c.active, False)
        for c in cdict.components():
            if c.storage_key == test_key:
                self.assertEqual(c.active, True)
            else:
                self.assertEqual(c.active, False)
        for c in cdict.components(active=True):
            self.assertEqual(c.active, True)
        self.assertNotEqual(
            len(list(cdict.components())), len(list(cdict.components(active=None)))
        )
        self.assertEqual(len(list(cdict.components(active=True))), 1)

    def test_preorder_traversal(self):
        cdict, traversal = super(
            _TestActiveDictContainerBase, self
        ).test_preorder_traversal()

        descend = lambda x: not x._is_heterogeneous_container

        cdict[1].deactivate()
        self.assertEqual(
            [None, '[0]', '[2]'],
            [
                c.name
                for c in pmo.preorder_traversal(cdict, active=True, descend=descend)
            ],
        )
        self.assertEqual(
            [id(cdict), id(cdict[0]), id(cdict[2])],
            [
                id(c)
                for c in pmo.preorder_traversal(cdict, active=True, descend=descend)
            ],
        )

        cdict[1].deactivate(shallow=False)
        self.assertEqual(
            [c.name for c in traversal if c.active],
            [
                c.name
                for c in pmo.preorder_traversal(cdict, active=True, descend=descend)
            ],
        )
        self.assertEqual(
            [id(c) for c in traversal if c.active],
            [
                id(c)
                for c in pmo.preorder_traversal(cdict, active=True, descend=descend)
            ],
        )

        cdict.deactivate()
        self.assertEqual(len(list(pmo.preorder_traversal(cdict, active=True))), 0)
        self.assertEqual(len(list(pmo.generate_names(cdict, active=True))), 0)

    def test_preorder_traversal_descend_check(self):
        cdict, traversal = super(
            _TestActiveDictContainerBase, self
        ).test_preorder_traversal_descend_check()

        cdict[1].deactivate()

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return not x._is_heterogeneous_container

        descend.seen = []
        # pmo.pprint(cdict)
        order = list(pmo.preorder_traversal(cdict, active=True, descend=descend))
        self.assertEqual([None, '[0]', '[2]'], [c.name for c in order])
        self.assertEqual(
            [id(cdict), id(cdict[0]), id(cdict[2])], [id(c) for c in order]
        )
        if cdict.ctype._is_heterogeneous_container:
            self.assertEqual([None, '[0]', '[2]'], [c.name for c in descend.seen])
            self.assertEqual(
                [id(cdict), id(cdict[0]), id(cdict[2])], [id(c) for c in descend.seen]
            )
        else:
            self.assertEqual([None], [c.name for c in descend.seen])
            self.assertEqual([id(cdict)], [id(c) for c in descend.seen])

        def descend(x):
            self.assertTrue(x._is_container)
            descend.seen.append(x)
            return x.active and (not x._is_heterogeneous_container)

        descend.seen = []
        order = list(pmo.preorder_traversal(cdict, active=None, descend=descend))
        self.assertEqual([None, '[0]', '[1]', '[2]'], [c.name for c in order])
        self.assertEqual(
            [id(cdict), id(cdict[0]), id(cdict[1]), id(cdict[2])],
            [id(c) for c in order],
        )
        if cdict.ctype._is_heterogeneous_container:
            self.assertEqual(
                [None, '[0]', '[1]', '[2]'], [c.name for c in descend.seen]
            )
            self.assertEqual(
                [id(cdict), id(cdict[0]), id(cdict[1]), id(cdict[2])],
                [id(c) for c in descend.seen],
            )
        else:
            self.assertEqual([None, '[1]'], [c.name for c in descend.seen])
            self.assertEqual([id(cdict), id(cdict[1])], [id(c) for c in descend.seen])

        cdict[1].deactivate(shallow=False)

        def descend(x):
            descend.seen.append(x)
            return not x._is_heterogeneous_container

        descend.seen = []
        order = list(pmo.preorder_traversal(cdict, active=True, descend=descend))
        self.assertEqual(
            [c.name for c in traversal if c.active], [c.name for c in order]
        )
        self.assertEqual([id(c) for c in traversal if c.active], [id(c) for c in order])
        self.assertEqual(
            [c.name for c in traversal if c.active and c._is_container],
            [c.name for c in descend.seen],
        )
        self.assertEqual(
            [id(c) for c in traversal if c.active and c._is_container],
            [id(c) for c in descend.seen],
        )

        def descend(x):
            descend.seen.append(x)
            return x.active and (not x._is_heterogeneous_container)

        descend.seen = []
        order = list(pmo.preorder_traversal(cdict, active=None, descend=descend))
        self.assertEqual([None, '[0]', '[1]', '[2]'], [c.name for c in order])
        self.assertEqual(
            [id(cdict), id(cdict[0]), id(cdict[1]), id(cdict[2])],
            [id(c) for c in order],
        )
        if cdict.ctype._is_heterogeneous_container:
            self.assertEqual(
                [None, '[0]', '[1]', '[2]'], [c.name for c in descend.seen]
            )
            self.assertEqual(
                [id(cdict), id(cdict[0]), id(cdict[1]), id(cdict[2])],
                [id(c) for c in descend.seen],
            )
        else:
            self.assertEqual([None, '[1]'], [c.name for c in descend.seen])
            self.assertEqual([id(cdict), id(cdict[1])], [id(c) for c in descend.seen])

        cdict.deactivate()

        def descend(x):
            descend.seen.append(x)
            return True

        descend.seen = []
        order = list(pmo.preorder_traversal(cdict, active=True, descend=descend))
        self.assertEqual(len(descend.seen), 0)
        self.assertEqual(len(list(pmo.generate_names(cdict, active=True))), 0)

        def descend(x):
            descend.seen.append(x)
            return x.active

        descend.seen = []
        order = list(pmo.preorder_traversal(cdict, active=None, descend=descend))
        self.assertEqual(len(descend.seen), 1)
        self.assertIs(descend.seen[0], cdict)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
import json

import pyomo.common.envvar as envvar
from pyomo.common.config import ConfigBase, ConfigBlock, ConfigValue, ADVANCED_OPTION
from pyomo.common.dependencies import yaml, yaml_available, yaml_load_args
import logging

logger = logging.getLogger('pyomo.core')


class _PyomoOptions:
    def __init__(self):
        self._options_stack = [default_pyomo_config()]

        # Load the user's configuration
        sources = [(json, 'json', True, 'json', {}), (json, 'jsn', True, 'json', {})]
        sources.append((yaml, 'yml', yaml_available, 'yaml', yaml_load_args))
        sources.append((yaml, 'yaml', yaml_available, 'yaml', yaml_load_args))
        for parser, suffix, available, library, parser_args in sources:
            cfg_file = os.path.join(envvar.PYOMO_CONFIG_DIR, 'config.' + suffix)
            if not os.path.exists(cfg_file):
                continue
            if not available:
                logger.warning(
                    "Default configuration file (%s) cannot be "
                    "loaded; %s is not available" % (cfg_file, library)
                )
                continue
            fp = open(cfg_file)
            try:
                data = parser.load(fp, **parser_args)
            except:
                logger.error(
                    "Error parsing the user's default "
                    "configuration file\n\t%s." % (cfg_file,)
                )
            self._options_stack[0].set_value(data)

    def active_config(self):
        return self._options_stack[-1]

    #
    # BEGIN clone of the ConfigBlock API
    #

    def __getitem__(self, key):
        return self.active_config().__getitem__(key)

    def get(self, key, default=ConfigBase.NoArgument):
        return self.active_config().get(key, default)

    def __setitem__(self, key, val):
        return self.active_config().__setitem__(key, val)

    def __contains__(self, key):
        return self.active_config().__contains__(key)

    def __len__(self):
        return self.active_config().__len__()

    def __iter__(self):
        return self.active_config().__iter__()

    def __getattr__(self, name):
        # if name in self.__dict__:
        #    return self.__dict__[name]
        return self._options_stack[-1].__getattr__(name)

    def __setattr__(self, name, value):
        if name == '_options_stack':
            super(_PyomoOptions, self).__setattr__(name, value)
        else:
            return self.active_config().__setattr__(name, value)

    def iterkeys(self):
        return self.active_config().iterkeys()

    def itervalues(self):
        return self.active_config().itervalues()

    def iteritems(self):
        return self.active_config().iteritems()

    #
    # Note: ConfigBlock correctly maps {keys,values,items} to
    # iter{keys,values,items} when running under Python3 ... so we
    # *don't* need to do that mapping here.
    #
    def keys(self):
        return self.active_config().keys()

    def values(self):
        return self.active_config().values()

    def items(self):
        return self.active_config().items()

    def declare(self, name, config):
        return self.active_config().items(name, config)

    def add(self, name, config):
        return self.active_config().add(name, config)

    def value(self, accessValue=True):
        return self.active_config().value(accessValue)

    def set_value(self, value):
        return self.active_config().set_value(value)

    def reset(self):
        return self.active_config().reset()

    #
    # END clone of the ConfigBlock API
    #


def default_pyomo_config():
    config = ConfigBlock("Pyomo configuration file")

    config.declare(
        'paranoia_level',
        ConfigValue(
            0,
            int,
            'Pyomo paranoia and error checking level',
            """Higher levels of paranoia enable additional error checking and
            warning messages that may assist users in identifying likely
            modeling problems.
            Default=0""",
            visibility=ADVANCED_OPTION,
        ),
    )

    return config


PyomoOptions = _PyomoOptions()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.kernel.base import (
    _no_ctype,
    _convert_ctype,
    _convert_descend_into,
    ICategorizedObjectContainer,
)


def heterogeneous_containers(node, ctype=_no_ctype, active=True, descend_into=True):
    """
    A generator that yields all heterogeneous containers
    included in an object storage tree, including the root
    object. Heterogeneous containers are categorized objects
    with a category type different from their children.

    Args:
        node: The root object.
        ctype: Indicates the category of objects to
            include. The default value indicates that
            all categories should be included.
        active (:const:`True`/:const:`None`): Controls
            whether or not to filter the iteration to
            include only the active part of the storage
            tree. The default is :const:`True`. Setting this
            keyword to :const:`None` causes the active
            status of objects to be ignored.
        descend_into (bool, function): Indicates whether or
            not to descend into a heterogeneous
            container. Default is True, which is equivalent
            to `lambda x: True`, meaning all heterogeneous
            containers will be descended into.

    Returns:
        iterator of heterogeneous containers in the storage
        tree, include the root object.
    """
    assert active in (None, True)

    # if not active, then nothing below is active
    if (active is not None) and (not node.active):
        return

    if not node.ctype._is_heterogeneous_container:
        return

    if not node._is_heterogeneous_container:
        assert node._is_container
        # a homogeneous container that stores
        # heterogeneous objects
        for obj in node.components(active=active):
            assert obj._is_heterogeneous_container
            yield from heterogeneous_containers(
                obj, ctype=ctype, active=active, descend_into=descend_into
            )
        return

    # convert AML types into Kernel types (hack for the
    # solver interfaces)
    ctype = _convert_ctype.get(ctype, ctype)
    assert (ctype is _no_ctype) or ctype._is_heterogeneous_container

    # convert descend_into to a function if
    # it is not already one
    descend_into = _convert_descend_into(descend_into)

    # a heterogeneous container
    if (ctype is _no_ctype) or (node.ctype is ctype):
        yield node

    if not descend_into(node):
        return

    for child_ctype in node.child_ctypes():
        if not child_ctype._is_heterogeneous_container:
            continue

        for child in node.children(ctype=child_ctype):
            assert child._is_container

            if (active is not None) and (not child.active):
                continue

            yield from heterogeneous_containers(
                child, ctype=ctype, active=active, descend_into=descend_into
            )


class IHeterogeneousContainer(ICategorizedObjectContainer):
    """
    A partial implementation of the ICategorizedObjectContainer
    interface for implementations that store multiple
    categories of objects.

    Complete implementations need to set the _ctype
    attribute and declare the remaining required abstract
    properties of the ICategorizedObjectContainer base
    class.
    """

    __slots__ = ()
    _is_heterogeneous_container = True

    #
    # Interface
    #

    def collect_ctypes(self, active=True, descend_into=True):
        """Returns the set of object category types that can
        be found under this container.

        Args:
            active (:const:`True`/:const:`None`): Controls
                whether or not to filter the iteration to
                include only the active part of the storage
                tree. The default is :const:`True`. Setting
                this keyword to :const:`None` causes the
                active status of objects to be ignored.
            descend_into (bool, function): Indicates whether
                or not to descend into a heterogeneous
                container. Default is True, which is
                equivalent to `lambda x: True`, meaning all
                heterogeneous containers will be descended
                into.

        Returns:
            A set of object category types
        """
        assert active in (None, True)

        ctypes = set()
        # if not active, then nothing below is active
        if (active is not None) and (not self.active):
            return ctypes

        # convert descend_into to a function if
        # it is not already one
        descend_into = _convert_descend_into(descend_into)

        for child_ctype in self.child_ctypes():
            for obj in self.components(
                ctype=child_ctype,
                active=active,
                descend_into=_convert_descend_into._false,
            ):
                ctypes.add(child_ctype)
                # just need 1 to appear in order to
                # count the child_ctype
                break

        if descend_into is _convert_descend_into._false:
            return ctypes

        # iterate over a copy of the set as we will
        # modify it in this loop
        for child_ctype in tuple(ctypes):
            if child_ctype._is_heterogeneous_container:
                for obj in self.components(
                    ctype=child_ctype,
                    active=active,
                    descend_into=_convert_descend_into._false,
                ):
                    assert obj._is_heterogeneous_container
                    if descend_into(obj):
                        ctypes.update(
                            obj.collect_ctypes(active=active, descend_into=descend_into)
                        )

        return ctypes

    def child_ctypes(self, *args, **kwds):
        """Returns the set of child object category types
        stored in this container."""
        raise NotImplementedError  # pragma:nocover

    #
    # Define the ICategorizedObjectContainer abstract methods
    #

    # def child(self, *args, **kwds):
    # ... not defined here

    # def children(self, *args, **kwds):
    # ... not defined here

    def components(self, ctype=_no_ctype, active=True, descend_into=True):
        """
        Generates an efficient traversal of all components
        stored under this container. Components are
        categorized objects that are either (1) not
        containers, or (2) are heterogeneous containers.

        Args:
            ctype: Indicates the category of components to
                include. The default value indicates that
                all categories should be included.
            active (:const:`True`/:const:`None`): Controls
                whether or not to filter the iteration to
                include only the active part of the storage
                tree. The default is :const:`True`. Setting
                this keyword to :const:`None` causes the
                active status of objects to be ignored.
            descend_into (bool, function): Indicates whether
                or not to descend into a heterogeneous
                container. Default is True, which is
                equivalent to `lambda x: True`, meaning all
                heterogeneous containers will be descended
                into.

        Returns:
            iterator of components in the storage tree
        """
        assert active in (None, True)

        # if not active, then nothing below is active
        if (active is not None) and (not self.active):
            return

        # convert AML types into Kernel types (hack for the
        # solver interfaces)
        ctype = _convert_ctype.get(ctype, ctype)

        # convert descend_into to a function if
        # it is not already one
        descend_into = _convert_descend_into(descend_into)

        if ctype is _no_ctype:
            for child in self.children():
                if (active is not None) and (not child.active):
                    continue

                if not child._is_container:
                    yield child
                elif child._is_heterogeneous_container:
                    yield child
                    if descend_into(child):
                        yield from child.components(
                            active=active, descend_into=descend_into
                        )
                elif (descend_into is _convert_descend_into._false) or (
                    not child.ctype._is_heterogeneous_container
                ):
                    assert child._is_container
                    yield from child.components(active=active)
                else:
                    assert child._is_container
                    for obj in child.components(active=active):
                        assert obj._is_heterogeneous_container
                        yield obj
                        if descend_into(obj):
                            yield from obj.components(
                                active=active, descend_into=descend_into
                            )

        else:
            for item in heterogeneous_containers(
                self, active=active, descend_into=descend_into
            ):
                for child in item.children(ctype=ctype):
                    if (not child._is_container) or child._is_heterogeneous_container:
                        if (active is None) or child.active:
                            yield child
                    else:
                        assert child._is_container
                        yield from child.components(active=active)
